/*
* JBoss, Home of Professional Open Source
* Copyright 2007, JBoss Inc., and individual contributors as indicated
* by the @authors tag. See the copyright.txt in the distribution for a
* full listing of individual contributors.
*
* This is free software; you can redistribute it and/or modify it
* under the terms of the GNU Lesser General Public License as
* published by the Free Software Foundation; either version 2.1 of
* the License, or (at your option) any later version.
*
* This software is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
* Lesser General Public License for more details.
*
* You should have received a copy of the GNU Lesser General Public
* License along with this software; if not, write to the Free
* Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
* 02110-1301 USA, or see the FSF site: http://www.fsf.org.
*/
package sample.vpd.ejb;

import java.sql.Connection;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;

import javax.annotation.Resource;
import javax.ejb.EJBContext;
import javax.ejb.Stateless;
import javax.sql.DataSource;

import org.jboss.logging.Logger;
import org.jboss.resource.adapter.jdbc.WrappedConnection;

/**
 * <p>A very simple business service, implemented as a stateless session bean which 
 * provides two business methods to query and update the balance of named 
 * bank accounts.</p>
 * 
 * <p>This bean is designed as a vehicle to demonstrate integration of Oracle VPD
 * and JBoss AS and manages its own access to a database via an injected 
 * DataSource object.</p>
 * 
 * <p>It does not enforce any security, nor is it aware of the fact that the 
 * underlying database supports Oracle VPD.</p>
 * 
 * @author <a href="pbennett@redhat.com">Pete Bennett</a>
 * @version $Revision: 1.1 $
 */
@Stateless
public class Bank
   implements BankRemote, BankLocal
{
   /** 
    * <p>The name of the table in the database which holds the account information.</p>
    * 
    * <p>This table is expected to be of the form:<br>
    * create table System.Accounts (Username varchar2(20), Balance float) tablespace SYSTEM;</p>
    */
   private static final String TABLE_NAME = "System.Accounts";
   
   /**
    * The fully qualified JNDI name for the injected DataSource
    */
   private static final String DATA_SOURCE_NAME = "java:VpdDS";
   
   /**
    * The error code corresponding to the database indicating the connected user has
    * insufficient privileges to execute the query they have attempted.
    */
   // TODO: find the correct constant from the Oracle Java classes rather than hard coding
   private static final int ORACLE_INSUFFICIENT_PRIVILEGES_ERROR_CODE = 1031;

   private Logger log = Logger.getLogger(Bank.class);

   @Resource
   EJBContext context;

   @Resource(mappedName = DATA_SOURCE_NAME)
   DataSource dataSource;

   /**
    * Queries a named bank account for its current balance.
    * 
    * @param accountName the name of the owner of the account
    * @return the current balance for the named account or null if the account 
    *         does not exist or the logged on user does not have permissions to
    *         view the balance
    */
   public Double getBalance(String accountName)
   {
      log.info("User " + getUser() + " called Bank.getBalance(accountName=" + accountName + ")");

      Object[] row = executeQuery("select Balance from " + TABLE_NAME + " where Username='" + accountName + "'");

      if (row == null)
      {
         return null;
      }
      else
      {
         return (Double) row[0];
      }
   }

   /**
    * Changes the balance for a named bank account.
    * 
    * @param accountName the name of the owner of the account
    * @param balance the new balance for the account
    * @throws SecurityException if the underlying database indicates the logged on user 
    *                           has insufficient privileges to modify the specified account's balance
    */
   public void setBalance(String accountName, Double balance)
   {
      log.info("User " + getUser() + " called Bank.setBalance(accountName=" + accountName + ", balance=" + balance
            + ")");

      //TODO: see comment on executeUpdate() method, this if statement clause does not neccesarily mean access has been denied at database level
      if (!executeUpdate("update " + TABLE_NAME + " set Balance = " + balance + " where Username = '" + accountName + "'"))
      {
         throw new SecurityException("Access Denied at Database Level");
      }
   }

   /**
    * Queries the EJBContext for this Entity and gets the Caller Principal, returning
    * the name associated with the prinicipal in the form of a String.
    * 
    * @return a String representation of the currently authenticated user or null if none
    */
   private String getUser()
   {
      return context.getCallerPrincipal().getName();
   }

   /**
    * Executes the specified query against a connection obtained from the injected DataSource
    * 
    * @param queryString the query String to be executed
    * @return an array of size one containing a single Double corresponding to the first element in the first row of the ResultSet returned
    *         from the database or null if no rows were returned
    */
   private Object[] executeQuery(String queryString)
   {
      log.info("executeQuery(" + queryString + ")");

      Connection connection = null;
      Statement statement = null;
      ResultSet resultSet = null;
      Object[] row = null;
      Connection underlyingConnection = null;

      try
      {
         connection = dataSource.getConnection();

         log.info("Connection Class=" + connection.getClass());

         statement = connection.createStatement();

         try
         {
            WrappedConnection wrappedConnection = (WrappedConnection) connection;

            underlyingConnection = wrappedConnection.getUnderlyingConnection();

            log.info("Underlying Connection Class=" + underlyingConnection.getClass());
         }
         catch (ClassCastException ex)
         {

         }

         resultSet = statement.executeQuery(queryString);

         row = new Object[1];

         if (resultSet.next())
         {
            row[0] = resultSet.getDouble(1);
         }
      }
      catch (SQLException ex)
      {
         log.error("Database Access Error [code=" + ex.getErrorCode() + "]", ex);

         row = null;
      }
      finally
      {
         if (statement != null)
         {
            try
            {
               statement.close();
            }
            catch (SQLException e)
            {
               log.error("Failed to close statement", e);
            }
         }

         if (connection != null)
         {
            try
            {
               connection.close();
            }
            catch (SQLException e)
            {
               log.error("Failed to close connection", e);
            }
         }
      }

      return row;
   }
   
   /**
    * Executes the specified update against a connection obtained from the injected DataSource
    * 
    * @param queryString the update String to be executed
    * @return true if the update was successful, false if it failed
    */
   //TODO: although you can see in the JBoss console whether the update failed due to VPD security
   //      or an unexpected error, the return value from this function does not reflect the 
   //      difference. This should be refactored.
   private boolean executeUpdate(String queryString)
   {
      log.info("executeUpdate(" + queryString + ")");

      Connection connection = null;
      Statement statement = null;
      boolean success = false;

      try
      {
         connection = dataSource.getConnection();

         statement = connection.createStatement();

         statement.executeUpdate(queryString);
         
         success = true;
      }
      catch (SQLException ex)
      {
         if (ex.getErrorCode() == ORACLE_INSUFFICIENT_PRIVILEGES_ERROR_CODE)
         {
            log.info("Access denied by Oracle security");
         }
         else
         {
            log.error("Database Access Error [code=" + ex.getErrorCode() + "]", ex);
         }
      }
      finally
      {
         if (statement != null)
         {
            try
            {
               statement.close();
            }
            catch (SQLException e)
            {
               log.error("Failed to close statement", e);
            }
         }

         if (connection != null)
         {
            try
            {
               connection.close();
            }
            catch (SQLException e)
            {
               log.error("Failed to close connection", e);
            }
         }
      }

      return success;
   }
}
